//
//  View.m
//  QueuePlay
//
//  Created by NYU User on 9/10/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import <AudioToolbox/AudioToolbox.h>
#include <mach/mach_time.h>	//for mach_absolute_time
#import "View.h"

#define KILOBYTE 1024	//number of bytes in a kilobyte
#define numberOutputBuffers ((size_t)3)

//Structure passed to outputCallback.

struct outputUserData {
	AudioStreamBasicDescription format;
	AudioQueueRef queue;
	AudioQueueBufferRef buffer[numberOutputBuffers];
	AudioFileID fileID;
	UInt32 bufferByteSize;
	SInt64 currentPacket;
	bool isRunning;
	
	UInt32 numPacketsToRead;
	AudioStreamPacketDescription *description;
};

static struct outputUserData outputData;

//An output audio queue calls the following function repeatedly
//to copy data from a file into a buffer.

static void outputCallback(
	void *userData,
	AudioQueueRef queue,
	AudioQueueBufferRef buffer
) {
	struct outputUserData *p = (struct outputUserData *)userData;
	if (!p->isRunning) {
		return;
	}
	
	UInt32 numBytesReadFromFile;
	UInt32 numPackets = p->numPacketsToRead;
	
	OSStatus status = AudioFileReadPackets(
		p->fileID,
		false,
		&numBytesReadFromFile,
		p->description,
		p->currentPacket,
		&numPackets,
		buffer->mAudioData
	);
	
	if (status != noErr) {
		NSLog(@"AudioFileReadPackets returned %d", status);
	}
	
	if (numPackets == 0) {
		status = AudioQueueStop(p->queue, false);
		if (status != noErr) {
			NSLog(@"AudioQueueStop returned %d", status);
		}
		p->isRunning = false;
	} else {
		buffer->mAudioDataByteSize = numBytesReadFromFile;
		AudioTimeStamp stamp;
		status = AudioQueueEnqueueBufferWithParameters(
			p->queue,
			buffer,
			p->description == NULL ? 0 : numPackets,
			p->description,
			0,
			0,
			0, NULL,
			NULL, &stamp
		);
		
		if (stamp.mFlags & kAudioTimeStampSampleTimeValid) {
			NSLog(@"stamp.mSampleTime / 21888 == %f",
				  stamp.mSampleTime / 21888);
		}
		
		if (stamp.mFlags & kAudioTimeStampHostTimeValid) {
			NSLog(@"stamp.mHostTime == %qu", stamp.mHostTime);
		}
		
		/*
		status = AudioQueueEnqueueBuffer(
			p->queue,
			buffer,
			p->description == NULL ? 0 : numPackets,
			p->description
		);
		 */
		if (status != noErr) {
			NSLog(@"outputCallback: AudioQueueEnqueueBuffer returned %d", status);
		}
		p->currentPacket += numPackets;
	}
}

static void outputBufferSize(
	AudioStreamBasicDescription *ASBDesc,
	UInt32 maxPacketSize,
	Float64 seconds,
	UInt32 *bufferSize,
	UInt32 *numPacketsToRead
) {
	static const UInt32 maxBufferSize = 320 * KILOBYTE;
	
	if (ASBDesc->mFramesPerPacket == 0) {
		*bufferSize = MAX(maxBufferSize, maxPacketSize);
	} else {
		*bufferSize = seconds * maxPacketSize * ASBDesc->mSampleRate
		/ ASBDesc->mFramesPerPacket;
	}
	
	if (*bufferSize > maxBufferSize
		&& *bufferSize > maxPacketSize) {
		*bufferSize = maxBufferSize;
	} else {
		*bufferSize = MAX(*bufferSize, 16 * KILOBYTE);
	}
	
	*numPacketsToRead = *bufferSize / maxPacketSize;
}

@implementation View

- (id) initWithFrame: (CGRect) frame {
	if ((self = [super initWithFrame:frame]) != nil) {
		// Initialization code
		self.backgroundColor = [UIColor whiteColor];
		
		NSBundle *bundle = [NSBundle mainBundle];
		NSString *fileName =
			[bundle pathForResource: @"musette" ofType: @"mp3"];
		
		NSURL *url = [NSURL fileURLWithPath: fileName isDirectory: NO];
		NSLog(@"url == \"%@\"", url);

		OSStatus status = AudioFileOpenURL(
			(CFURLRef)url,
			kAudioFileReadPermission,
			kAudioFileMP3Type,
			&outputData.fileID
		);
		if (status != noErr) {
			NSLog(@"AudioFileOpenURL returned %d", status);
		}
		
		UInt32 size = sizeof outputData.format;
		status = AudioFileGetProperty(
			outputData.fileID,
			kAudioFilePropertyDataFormat,
			&size,
			&outputData.format
		);
		if (status != noErr) {
			NSLog(@"AudioFileGetProperty", status);
		}
		//NSLog(@"size == %u", size);
		
		status = AudioQueueNewOutput(
			&outputData.format,
			outputCallback,
			&outputData,
			CFRunLoopGetCurrent(),
			kCFRunLoopCommonModes,
			0,
			&outputData.queue
		);
		if (status != noErr) {
			NSLog(@"AudioQueueNewOutput returned %d", status);
		}
		
		UInt32 maxPacketSize;
		UInt32 propertySize = sizeof maxPacketSize;
		
		status = AudioFileGetProperty(
			outputData.fileID,
			kAudioFilePropertyPacketSizeUpperBound,
			&propertySize,
			&maxPacketSize
		);
		if (status != noErr) {
			NSLog(@"AudioFileGetProperty returned %d", status);
		}
		//NSLog(@"maxPacketSize == %u", maxPacketSize);
		
		outputBufferSize(
			&outputData.format,
			maxPacketSize,
			.5,
			&outputData.bufferByteSize,
			&outputData.numPacketsToRead
		);
		
		NSLog(@"outputData.bufferByteSize == %u",
			  outputData.bufferByteSize);
		NSLog(@"outputData.numPacketsToRead == %u",
			  outputData.numPacketsToRead);
		
		//constant bitrate
		bool isCBR = outputData.format.mBytesPerPacket > 0
		&& outputData.format.mFramesPerPacket > 0;
		
		if (isCBR) {
			outputData.description = NULL;
		} else {
			outputData.description = (AudioStreamPacketDescription *)
			malloc(outputData.numPacketsToRead
				* sizeof (AudioStreamPacketDescription));
			
			if (outputData.description == NULL) {
				NSLog(@"malloc failed");
			}
		}
		
		//Prime the queue with some data before starting.
		outputData.currentPacket = 0;
		outputData.isRunning = true;
		
		for (size_t i = 0; i < numberOutputBuffers; ++i) {
			status = AudioQueueAllocateBuffer(
				outputData.queue,
				outputData.bufferByteSize,
				outputData.buffer + i
			);
			
			if (status != noErr) {
				NSLog(@"AudioQueueAllocateBuffer returned %d", status);
			}
			
			outputCallback(
				&outputData,
				outputData.queue,
				outputData.buffer[i]
			);
		}
		
		status = AudioQueueSetParameter(
			outputData.queue,
			kAudioQueueParam_Volume,
			1.0
		);
		
		if (status != noErr) {
			NSLog(@"AudioQueueSetParameter returned %d", status);
		}
		
		AudioTimeStamp stamp = {0};
		stamp.mFlags = kAudioTimeStampHostTimeValid;
		
		mach_timebase_info_data_t info;
		mach_timebase_info(&info);
		
		stamp.mHostTime = mach_absolute_time() +
			10 * 1000000000.0
			* info.numer / info.denom;
		NSLog(@"stamp.mHostTime == %qu", stamp.mHostTime);
		NSLog(@"info.numer == %u", info.numer);
		NSLog(@"info.denom == %u", info.denom);
		status = AudioQueueStart(outputData.queue, &stamp);
		if (status != noErr) {
			NSLog(@"AudioQueueStart returned %d", status);
		}
		
		do {
			SInt32 result = CFRunLoopRunInMode(
				kCFRunLoopDefaultMode,
				.25,
				false
			);
			
			if (result != kCFRunLoopRunTimedOut) {
				NSLog(@"CFRunLoopRunInMode returned %d", result);
			}
		} while (outputData.isRunning);
		
		SInt32 result = CFRunLoopRunInMode(
			kCFRunLoopDefaultMode,
			1.0,
			false
		);
		
		if (result != kCFRunLoopRunTimedOut) {
			NSLog(@"CFRunLoopRunInMode returned %d", result);
		}
		
	}
	return self;
}

/*
// Only override drawRect: if you perform custom drawing.
// An empty implementation adversely affects performance during animation.
- (void) drawRect: (CGRect) rect {
	// Drawing code
}
*/

- (void) dealloc {
	[super dealloc];
}

@end
